"""
2021 Sebastian de Bone (QuTech)
https://github.com/sebastiandebone/ghz_prot_II
"""
import numpy as np

import da_search as das
import da_protocols as dap
import simulate_protocol.protocol_recipe as pr
from circuit_simulation.stabilizer_measurement_protocols.run_protocols import main


def minimum_GHZ_creation_time(GHZ_target_fidelity=0.8, data=None, da_type='mpF',  n_max=4, k_max=8,
                              Bell_pair_fidelity=0.9, number_of_state_stored=3, use_circuit_simulator=False,
                              number_of_iterations_simulator=1):
    if data is None:
        data, _ = das.dynamic_algorithm(n_max, k_max, Bell_pair_fidelity, da_type=da_type, nstate=number_of_state_stored,
                                        rot_type="none", show_or_not=1, calculate_prot=use_circuit_simulator)

    smallest_time = [np.inf] * (n_max - 1)

    arguments = {'color': False, 'save_latex_pdf': False, 'csv_filename': None, 'cp_path': None, 'threaded': False,
                 'to_console': False, 'draw_circuit': False, 'single_qubit_gate_lookup': None,
                 'two_qubit_gate_lookup': None, 'protocol': "auto_generated_swap",
                 'gate_duration_file': 'circuit_simulation\\gate_duration.txt',
                 'progress_bar': False, 'cut_off_file': None,
                 'force_run': True, 'pg': 0.001, 'pm': 0.001, 'pm_1': 0.001, 'pn': 0.013, 'lde_success': 0.01,
                 'fixed_lde_attempts': 100, 'pulse_duration': 13e-3, '_node': 'Pur', 'cut_off_time': np.inf,
                 'iterations': number_of_iterations_simulator, 'stabilizer_type': 'Z', 'decoherence': True,
                 'pm_equals_pg': True,
                 'probabilistic': True, 'measurement_duration': 1e-06, 'lde_duration': 1e-06,
                 'use_swap_gates': True,
                 'noiseless_swap': False, 'bell_pair_type': 3, 'network_noise_type': 0, 'T1_idle': 300.0,
                 'T2_idle': 10.0, 'T1_lde': 1.2, 'T2_lde': 1.2, 'T1_idle_electron': 300, 'T2_idle_electron': 1.0,
                 'combine': True, 'cutoff_search': False, 'seed_number': None}

    for n in range(2, n_max + 1):
        for k in range(n - 1, len(data[n])):
            for b in range(len(data[n][k])):
                protocol = dap.identify_protocol(data, n, k, b)
                protocol_recipe = pr.ProtocolRecipe(protocol)
                _, _, char = main(**arguments, protocol_recipe=protocol_recipe)
                average_ghz_fid = np.average(char['ghz_fid'])
                if average_ghz_fid >= GHZ_target_fidelity:
                    average_time = np.average(char['dur'])
                    if average_time < smallest_time[n - 2]:
                        smallest_time[n - 2] = average_time

    return smallest_time