import numpy as np
from math import ceil
import multiprocessing
import time
from copy import deepcopy

from circuit_simulation.stabilizer_measurement_protocols.run_protocols import main


def run_circuit_simulator_threaded():
    threads = multiprocessing.cpu_count()
    pool = multiprocessing.Pool(threads)

    print(f"Started calculating with {threads} threads.")
    tic = time.perf_counter()

    iterations_per_thread = 6000
    offset = 128 * 300

    results = []
    for i in range(threads):
        results.append(pool.apply_async(run_circuit_simulator,
                                        args=[[*range(offset + iterations_per_thread*i,
                                                      offset + iterations_per_thread*(i+1))],
                                              'dyn_prot_4_4_1_swap',
                                              None]))

    low_seeds_total = []
    cut_off_reached_total = []
    for i, result in enumerate(results):
        low_seeds, cut_off_reached_seeds = result.get()
        for low_seed in low_seeds:
            low_seeds_total.append(low_seed)
        for cut_off_reached_seed in cut_off_reached_seeds:
            cut_off_reached_total.append(cut_off_reached_seed)

    pool.close()

    toc = time.perf_counter()
    print(f"Finished calculating in {toc - tic:0.4f} seconds.")
    print(f"Low seeds: {low_seeds_total}.")
    print(f"Cut off reached seeds: {cut_off_reached_total}.")


def run_circuit_simulator(seeds=None, protocol_name='dyn_prot_4_4_1_swap', protocol_recipe=None):
    numb_of_iter = 2
    arguments = {'color': False, 'save_latex_pdf': False, 'csv_filename': None, 'cp_path': None, 'threaded': True,
                 'to_console': False, 'draw_circuit': False, 'single_qubit_gate_lookup': None,
                 'thread_safe_printing': False,
                 'two_qubit_gate_lookup': None, 'protocol': protocol_name,
                 # 'gate_duration_file': '..\\..\\Paul\\circuit_simulation\\gate_duration.txt',
                 'progress_bar': False, 'cut_off_file': None, 'force_run': True, 'pg': 0.000175889, 'pm': 0.000175889,
                 'pm_1': 0.000175889,
                 'pn': 0.015, 'lde_success': 0.017, 'fixed_lde_attempts': 100, 'pulse_duration': 13e-3, '_node': 'Pur',
                 'cut_off_time': 0.3023, 'iterations': numb_of_iter, 'stabilizer_type': 'Z', 'decoherence': True,
                 'pm_equals_pg': True, 'probabilistic': True, 'measurement_duration': 1e-06, 'lde_duration': 1e-06,
                 'noiseless_swap': False, 'bell_pair_type': 3, 'network_noise_type': 3, 'T1_idle': 300,
                 'T2_idle': 10.0, 'T1_lde': 1.2, 'T2_lde': 1.2, 'T1_idle_electron': 300, 'T2_idle_electron': 1.0,
                 'combine': True, 'protocol_recipe': protocol_recipe, 'seed_number': None,
                 'calc_id': 1, 'iter_pw': 3120}

    if not isinstance(seeds, list):
        seeds = [seeds]
    print(f"Circuit simulator is running {len(seeds)} times with {numb_of_iter} iterations, for seed = {seeds[0]} to seed = {seeds[-1]}.")
    low_seeds = []
    cut_off_reached_seeds =[]
    for seed_nmb in seeds:
        # arguments['seed_number'] = seed_nmb
        arguments_copy = deepcopy(arguments)
        (succeeded, failed), print_lines, characteristics = main(**arguments_copy)
        print("\n\n\n\n\n")
        print(characteristics['stab_fid'])
        print(characteristics['ghz_fid'])
        # a = characteristics['stab_fid']
        # if a:
        #     b = a[0]
        #     if b < 0.7:
        #         low_seeds.append(seed_nmb)
        # else:
        #     cut_off_reached_seeds.append(seed_nmb)

    # print(*print_lines)
    # print(succeeded)
    return low_seeds, cut_off_reached_seeds


if __name__ == "__main__":
    run_circuit_simulator()