from circuit_simulation.circuit_simulator import *
# from circuit_simulation.stabilizer_measurement_protocols.run_protocols import add_decoherence_if_cut_off, \
#     SUM_ACCURACY, _combine_idle_and_stabilizer_superoperator, _combine_superoperator_dataframes
from math import ceil
import sys

if __name__ == "__main__":

    arguments = {'color': False, 'save_latex_pdf': False,
                 'to_console': False, 'draw_circuit': False, 'single_qubit_gate_lookup': None,
                 'two_qubit_gate_lookup': None, 'stabilizer_type': 'Z',
                 'force_run': True,
                 'pg': 0.001,
                 'pm': 0.001,
                 'pm_1': 0.001,
                 'pn': 0.024,
                 'lde_success': 0.00414,
                 'fixed_lde_attempts': 0,
                 'pulse_duration': 0,
                 'cut_off_time': np.inf,
                 'decoherence': True,
                 'pm_equals_pg': True,
                 'probabilistic': False,
                 'measurement_duration': 1e-06,
                 'lde_duration': 1e-06,
                 'noiseless_swap': False,
                 'bell_pair_type': 3,
                 'network_noise_type': 2,
                 'T1_idle': 300.0,
                 'T2_idle': 10.0,
                 'T1_lde': 10,
                 'T2_lde': 10,
                 'T1_idle_electron': 300,
                 'T2_idle_electron': 1.0,
                 'noise': True, 'basis_transformation_noise': False, 'no_single_qubit_error': True}

    average_link_attempts = int(1 / arguments["lde_success"])

    supop_dataframe_failed = None
    supop_dataframe_succeed = None
    total_print_lines = []
    characteristics = {'dur': [], 'stab_fid': [], 'ghz_fid': []}

    operation = CZ_gate if arguments["stabilizer_type"] == "Z" else CNOT_gate

    qc = QuantumCircuit(8, 2, **arguments)

    qc.define_node("A", qubits=[6, 3, 2])
    qc.define_node("B", qubits=[4, 1, 0])
    qc.define_sub_circuit("AB")
    qc.define_sub_circuit("A")
    qc.define_sub_circuit("B", concurrent_sub_circuits="A")

    qc.start_sub_circuit("AB")
    qc.create_bell_pair("A-e", "B-e", attempts=average_link_attempts)
    qc.stabilizer_measurement(operation, nodes=["B", "A"])
    qc.end_current_sub_circuit(total=True, forced_level=True, apply_decoherence=True)
    superoperator_qubits = qc.data_qubits
    _, dataframe = qc.get_superoperator(superoperator_qubits, arguments["stabilizer_type"],
                                        no_color=(not arguments["color"]), stabilizer_protocol=True,
                                        print_to_console=arguments["to_console"], idle_data_qubit=False,
                                        protocol_name="Distillation_regime")
    direct_stab_fid = dataframe.iloc[0, 0]

    number_of_distillation_steps = 1
    stab_fid = [1]
    ghz_fid = []
    while number_of_distillation_steps < 100 and stab_fid[-1] > direct_stab_fid:
        distillation_gate = [CZ_gate, CNOT_gate, CiY_gate] * ceil(number_of_distillation_steps / 3)
        qc.start_sub_circuit("AB")
        qc.create_bell_pair("A-e", "B-e", attempts=average_link_attempts)
        qc.SWAP("A-e", "A-e+1")
        qc.SWAP("B-e", "B-e+1")
        for i in range(number_of_distillation_steps):
            gate = distillation_gate[i]
            qc.create_bell_pair("A-e", "B-e", attempts=average_link_attempts)
            qc.apply_gate(gate, cqubit="A-e", tqubit="A-e+1")
            qc.apply_gate(gate, cqubit="B-e", tqubit="B-e+1")
            qc.measure(["B-e", "A-e"], basis="X")

        qc.stabilizer_measurement(operation, nodes=["B", "A"])

        qc.end_current_sub_circuit(total=True, forced_level=True, apply_decoherence=True)

        qc.draw_circuit(no_color=not arguments["color"], color_nodes=True) if arguments["draw_circuit"] else None
        qc.draw_circuit_latex() if arguments["save_latex_pdf"] else None

        superoperator_qubits = qc.data_qubits

        # Obtain the superoperator in a dataframe format
        idle_data_qubit = False
        _, dataframe = qc.get_superoperator(superoperator_qubits, arguments["stabilizer_type"],
                                            no_color=(not arguments["color"]), stabilizer_protocol=True,
                                            print_to_console=arguments["to_console"], idle_data_qubit=idle_data_qubit,
                                            protocol_name="Distillation_regime")

        stab_fid.append(dataframe.iloc[0, 0])
        ghz_fid.append(qc.ghz_fidelity)
        number_of_distillation_steps += 1

    characteristics['dur'] += [qc.total_duration]
    characteristics['ghz_fid'] += [qc.ghz_fidelity]
    characteristics['stab_fid'] += [dataframe.iloc[0, 0]]

    total_print_lines.extend(qc.print_lines)
    print(direct_stab_fid)
    print(number_of_distillation_steps)
    print(stab_fid)
    print(ghz_fid)
    # print(qc.ghz_fidelity)
    # print(characteristics['stab_fid'][0])
    # print(characteristics["dur"][0])

